package org.eclipse.jdt.internal.compiler.ast;

import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.lookup.ImportBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.CompilationUnitScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;

public class Javadoc extends ASTNode
{
    public JavadocSingleNameReference[] paramReferences;
    public JavadocSingleTypeReference[] paramTypeParameters;
    public TypeReference[] exceptionReferences;
    public JavadocReturnStatement returnStatement;
    public Expression[] seeReferences;
    public long[] inheritedPositions;
    public JavadocSingleNameReference[] invalidParameters;
    public long valuePositions;
    
    public Javadoc(int sourceStart, int sourceEnd) {
        this.inheritedPositions = null;
        this.valuePositions = -1L;
        this.sourceStart = sourceStart;
        this.sourceEnd = sourceEnd;
        this.bits |= 0x10000;
    }
    
    boolean canBeSeen(int visibility, int modifiers) {
        if (modifiers < 0) {
            return true;
        }
        switch (modifiers & 0x7) {
            case 1: {
                return true;
            }
            case 4: {
                return visibility != 1;
            }
            case 0: {
                return visibility == 0 || visibility == 2;
            }
            case 2: {
                return visibility == 2;
            }
            default: {
                return true;
            }
        }
    }
    
    public ASTNode getNodeStartingAt(int start) {
        int length = 0;
        if (this.paramReferences != null) {
            length = this.paramReferences.length;
            int i = 0;
            while (i < length) {
                JavadocSingleNameReference param = this.paramReferences[i];
                if (param.sourceStart == start) {
                    return param;
                }
                ++i;
            }
        }
        if (this.invalidParameters != null) {
            length = this.invalidParameters.length;
            int i = 0;
            while (i < length) {
                JavadocSingleNameReference param = this.invalidParameters[i];
                if (param.sourceStart == start) {
                    return param;
                }
                ++i;
            }
        }
        if (this.paramTypeParameters != null) {
            length = this.paramTypeParameters.length;
            int i = 0;
            while (i < length) {
                JavadocSingleTypeReference param2 = this.paramTypeParameters[i];
                if (param2.sourceStart == start) {
                    return param2;
                }
                ++i;
            }
        }
        if (this.exceptionReferences != null) {
            length = this.exceptionReferences.length;
            int i = 0;
            while (i < length) {
                TypeReference typeRef = this.exceptionReferences[i];
                if (typeRef.sourceStart == start) {
                    return typeRef;
                }
                ++i;
            }
        }
        if (this.seeReferences != null) {
            length = this.seeReferences.length;
            int i = 0;
            while (i < length) {
                Expression expression = this.seeReferences[i];
                if (expression.sourceStart == start) {
                    return expression;
                }
                if ((expression instanceof JavadocAllocationExpression)) {
                    JavadocAllocationExpression allocationExpr = (JavadocAllocationExpression)this.seeReferences[i];
                    if (allocationExpr.binding != null && allocationExpr.binding.isValidBinding() && allocationExpr.arguments != null) {
                        int j = 0;
                        int l = allocationExpr.arguments.length;
                        while (j < l) {
                            if (allocationExpr.arguments[j].sourceStart == start) {
                                return allocationExpr.arguments[j];
                            }
                            ++j;
                        }
                    }
                }
                else if ((expression instanceof JavadocMessageSend)) {
                    JavadocMessageSend messageSend = (JavadocMessageSend)this.seeReferences[i];
                    if (messageSend.binding != null && messageSend.binding.isValidBinding() && messageSend.arguments != null) {
                        int j = 0;
                        int l = messageSend.arguments.length;
                        while (j < l) {
                            if (messageSend.arguments[j].sourceStart == start) {
                                return messageSend.arguments[j];
                            }
                            ++j;
                        }
                    }
                }
                ++i;
            }
        }
        return null;
    }
    
    @Override
    public StringBuffer print(int indent, StringBuffer output) {
        ASTNode.printIndent(indent, output).append("/**\n");
        if (this.paramReferences != null) {
            for (int i = 0, length = this.paramReferences.length; i < length; ++i) {
                ASTNode.printIndent(indent + 1, output).append(" * @param ");
                this.paramReferences[i].print(indent, output).append('\n');
            }
        }
        if (this.paramTypeParameters != null) {
            for (int i = 0, length = this.paramTypeParameters.length; i < length; ++i) {
                ASTNode.printIndent(indent + 1, output).append(" * @param <");
                this.paramTypeParameters[i].print(indent, output).append(">\n");
            }
        }
        if (this.returnStatement != null) {
            ASTNode.printIndent(indent + 1, output).append(" * @");
            this.returnStatement.print(indent, output).append('\n');
        }
        if (this.exceptionReferences != null) {
            for (int i = 0, length = this.exceptionReferences.length; i < length; ++i) {
                ASTNode.printIndent(indent + 1, output).append(" * @throws ");
                this.exceptionReferences[i].print(indent, output).append('\n');
            }
        }
        if (this.seeReferences != null) {
            for (int i = 0, length = this.seeReferences.length; i < length; ++i) {
                ASTNode.printIndent(indent + 1, output).append(" * @see ");
                this.seeReferences[i].print(indent, output).append('\n');
            }
        }
        ASTNode.printIndent(indent, output).append(" */\n");
        return output;
    }
    
    public void resolve(ClassScope scope) {
        if ((this.bits & 0x10000) == 0x0) {
            return;
        }
        if (this.inheritedPositions != null) {
            for (int length = this.inheritedPositions.length, i = 0; i < length; ++i) {
                int start = (int)(this.inheritedPositions[i] >>> 32);
                int end = (int)this.inheritedPositions[i];
                scope.problemReporter().javadocUnexpectedTag(start, end);
            }
        }
        for (int paramTagsSize = (this.paramReferences == null) ? 0 : this.paramReferences.length, i = 0; i < paramTagsSize; ++i) {
            JavadocSingleNameReference param = this.paramReferences[i];
            scope.problemReporter().javadocUnexpectedTag(param.tagSourceStart, param.tagSourceEnd);
        }
        this.resolveTypeParameterTags(scope, true);
        if (this.returnStatement != null) {
            scope.problemReporter().javadocUnexpectedTag(this.returnStatement.sourceStart, this.returnStatement.sourceEnd);
        }
        for (int throwsTagsLength = (this.exceptionReferences == null) ? 0 : this.exceptionReferences.length, j = 0; j < throwsTagsLength; ++j) {
            TypeReference typeRef = this.exceptionReferences[j];
            int start2;
            int end2;
            if ((typeRef instanceof JavadocSingleTypeReference)) {
                JavadocSingleTypeReference singleRef = (JavadocSingleTypeReference)typeRef;
                start2 = singleRef.tagSourceStart;
                end2 = singleRef.tagSourceEnd;
            }
            else if ((typeRef instanceof JavadocQualifiedTypeReference)) {
                JavadocQualifiedTypeReference qualifiedRef = (JavadocQualifiedTypeReference)typeRef;
                start2 = qualifiedRef.tagSourceStart;
                end2 = qualifiedRef.tagSourceEnd;
            }
            else {
                start2 = typeRef.sourceStart;
                end2 = typeRef.sourceEnd;
            }
            scope.problemReporter().javadocUnexpectedTag(start2, end2);
        }
        for (int seeTagsLength = (this.seeReferences == null) ? 0 : this.seeReferences.length, k = 0; k < seeTagsLength; ++k) {
            this.resolveReference(this.seeReferences[k], scope);
        }
        boolean source15 = scope.compilerOptions().sourceLevel >= 3211264L;
        if (!source15 && this.valuePositions != -1L) {
            scope.problemReporter().javadocUnexpectedTag((int)(this.valuePositions >>> 32), (int)this.valuePositions);
        }
        return;
    }
    
    public void resolve(CompilationUnitScope unitScope) {
        if ((this.bits & 0x10000) == 0x0) {
            return;
        }
        return;
    }
    
    public void resolve(MethodScope methScope) {
        if ((this.bits & 0x10000) == 0x0) {
            return;
        }
        AbstractMethodDeclaration methDecl = methScope.referenceMethod();
        boolean overriding = methDecl != null && methDecl.binding != null && (!methDecl.binding.isStatic() && (methDecl.binding.modifiers & 0x30000000) != 0x0);
        int seeTagsLength = (this.seeReferences == null) ? 0 : this.seeReferences.length;
        boolean superRef = false;
        for (int i = 0; i < seeTagsLength; ++i) {
            this.resolveReference(this.seeReferences[i], methScope);
            if (methDecl != null && !superRef) {
                if (!methDecl.isConstructor()) {
                    if (overriding && this.seeReferences[i] instanceof JavadocMessageSend) {
                        JavadocMessageSend messageSend = (JavadocMessageSend)this.seeReferences[i];
                        if (messageSend.binding != null && messageSend.binding.isValidBinding() && messageSend.actualReceiverType instanceof ReferenceBinding) {
                            ReferenceBinding methodReceiverType = (ReferenceBinding)messageSend.actualReceiverType;
                            TypeBinding superType2 = methDecl.binding.declaringClass.findSuperTypeOriginatingFrom(methodReceiverType);
                            if (superType2 != null && TypeBinding.notEquals(superType2.original(), methDecl.binding.declaringClass) && CharOperation.equals(messageSend.selector, methDecl.selector) && methScope.environment().methodVerifier().doesMethodOverride(methDecl.binding, messageSend.binding.original())) {
                                superRef = true;
                            }
                        }
                    }
                }
                else if ((this.seeReferences[i] instanceof JavadocAllocationExpression)) {
                    JavadocAllocationExpression allocationExpr = (JavadocAllocationExpression)this.seeReferences[i];
                    if (allocationExpr.binding != null && allocationExpr.binding.isValidBinding()) {
                        ReferenceBinding allocType = (ReferenceBinding)allocationExpr.resolvedType.original();
                        ReferenceBinding superType = (ReferenceBinding)methDecl.binding.declaringClass.findSuperTypeOriginatingFrom(allocType);
                        if (superType != null && TypeBinding.notEquals(superType.original(), methDecl.binding.declaringClass)) {
                            MethodBinding superConstructor = methScope.getConstructor(superType, methDecl.binding.parameters, allocationExpr);
                            if (superConstructor.isValidBinding() && superConstructor.original() == allocationExpr.binding.original()) {
                                MethodBinding current = methDecl.binding;
                                if (methScope.compilerOptions().sourceLevel >= 3407872L && current.typeVariables != Binding.NO_TYPE_VARIABLES) {
                                    current = current.asRawMethod(methScope.environment());
                                }
                                if (superConstructor.areParametersEqual(current)) {
                                    superRef = true;
                                }
                            }
                        }
                    }
                }
            }
        }
        if (!superRef && methDecl != null && methDecl.annotations != null) {
            for (int length = methDecl.annotations.length, j = 0; j < length && !superRef; superRef = ((methDecl.binding.tagBits & 0x2000000000000L) != 0x0L), ++j) {}
        }
        boolean reportMissing = methDecl == null || ((!overriding || this.inheritedPositions == null) && !superRef && (methDecl.binding.declaringClass == null || !methDecl.binding.declaringClass.isLocalType()));
        if (!overriding && this.inheritedPositions != null) {
            for (int length2 = this.inheritedPositions.length, k = 0; k < length2; ++k) {
                int start = (int)(this.inheritedPositions[k] >>> 32);
                int end = (int)this.inheritedPositions[k];
                methScope.problemReporter().javadocUnexpectedTag(start, end);
            }
        }
        CompilerOptions compilerOptions = methScope.compilerOptions();
        this.resolveParamTags(methScope, reportMissing, compilerOptions.reportUnusedParameterIncludeDocCommentReference);
        this.resolveTypeParameterTags(methScope, reportMissing && compilerOptions.reportMissingJavadocTagsMethodTypeParameters);
        if (this.returnStatement == null) {
            if (reportMissing && methDecl != null && methDecl.isMethod()) {
                MethodDeclaration meth = (MethodDeclaration)methDecl;
                if (meth.binding.returnType != TypeBinding.VOID) {
                    methScope.problemReporter().javadocMissingReturnTag(meth.returnType.sourceStart, meth.returnType.sourceEnd, methDecl.binding.modifiers);
                }
            }
        }
        else {
            this.returnStatement.resolve(methScope);
        }
        this.resolveThrowsTags(methScope, reportMissing);
        boolean source15 = compilerOptions.sourceLevel >= 3211264L;
        if (!source15 && methDecl != null && this.valuePositions != -1L) {
            methScope.problemReporter().javadocUnexpectedTag((int)(this.valuePositions >>> 32), (int)this.valuePositions);
        }
        for (int length3 = (this.invalidParameters == null) ? 0 : this.invalidParameters.length, l = 0; l < length3; ++l) {
            this.invalidParameters[l].resolve(methScope, false, false);
        }
        return;
    }
    
    private void resolveReference(Expression reference, Scope scope) {
        int problemCount = scope.referenceContext().compilationResult().problemCount;
        switch (scope.kind) {
            case 2: {
                reference.resolveType((MethodScope)scope);
                break;
            }
            case 3: {
                reference.resolveType((ClassScope)scope);
                break;
            }
        }
        boolean hasProblems = scope.referenceContext().compilationResult().problemCount > problemCount;
        boolean source15 = scope.compilerOptions().sourceLevel >= 3211264L;
        int scopeModifiers = -1;
        if ((reference instanceof JavadocFieldReference)) {
            JavadocFieldReference fieldRef = (JavadocFieldReference)reference;
            if (fieldRef.methodBinding != null) {
                if (fieldRef.tagValue == 10) {
                    if (scopeModifiers == -1) {
                        scopeModifiers = scope.getDeclarationModifiers();
                    }
                    scope.problemReporter().javadocInvalidValueReference(fieldRef.sourceStart, fieldRef.sourceEnd, scopeModifiers);
                }
                else if (fieldRef.actualReceiverType != null) {
                    if (scope.enclosingSourceType().isCompatibleWith(fieldRef.actualReceiverType)) {
                        fieldRef.bits |= 0x4000;
                    }
                    ReferenceBinding resolvedType2 = (ReferenceBinding)fieldRef.actualReceiverType;
                    if (CharOperation.equals(resolvedType2.sourceName(), fieldRef.token)) {
                        fieldRef.methodBinding = scope.getConstructor(resolvedType2, Binding.NO_TYPES, fieldRef);
                    }
                    else {
                        fieldRef.methodBinding = scope.findMethod(resolvedType2, fieldRef.token, Binding.NO_TYPES, fieldRef, false);
                    }
                }
            }
            else if (source15 && fieldRef.binding != null && fieldRef.binding.isValidBinding() && fieldRef.tagValue == 10 && !fieldRef.binding.isStatic()) {
                if (scopeModifiers == -1) {
                    scopeModifiers = scope.getDeclarationModifiers();
                }
                scope.problemReporter().javadocInvalidValueReference(fieldRef.sourceStart, fieldRef.sourceEnd, scopeModifiers);
            }
            if (!hasProblems && fieldRef.binding != null && fieldRef.binding.isValidBinding() && fieldRef.actualReceiverType instanceof ReferenceBinding) {
                ReferenceBinding resolvedType2 = (ReferenceBinding)fieldRef.actualReceiverType;
                this.verifyTypeReference(fieldRef, fieldRef.receiver, scope, source15, resolvedType2, fieldRef.binding.modifiers);
            }
            return;
        }
        else {
            if (!hasProblems && (reference instanceof JavadocSingleTypeReference || reference instanceof JavadocQualifiedTypeReference) && reference.resolvedType instanceof ReferenceBinding) {
                ReferenceBinding resolvedType = (ReferenceBinding)reference.resolvedType;
                this.verifyTypeReference(reference, reference, scope, source15, resolvedType, resolvedType.modifiers);
            }
            if ((reference instanceof JavadocMessageSend)) {
                JavadocMessageSend msgSend = (JavadocMessageSend)reference;
                if (source15 && msgSend.tagValue == 10) {
                    if (scopeModifiers == -1) {
                        scopeModifiers = scope.getDeclarationModifiers();
                    }
                    scope.problemReporter().javadocInvalidValueReference(msgSend.sourceStart, msgSend.sourceEnd, scopeModifiers);
                }
                if (!hasProblems && msgSend.binding != null && msgSend.binding.isValidBinding() && msgSend.actualReceiverType instanceof ReferenceBinding) {
                    ReferenceBinding resolvedType2 = (ReferenceBinding)msgSend.actualReceiverType;
                    this.verifyTypeReference(msgSend, msgSend.receiver, scope, source15, resolvedType2, msgSend.binding.modifiers);
                }
            }
            else if ((reference instanceof JavadocAllocationExpression)) {
                JavadocAllocationExpression alloc = (JavadocAllocationExpression)reference;
                if (source15 && alloc.tagValue == 10) {
                    if (scopeModifiers == -1) {
                        scopeModifiers = scope.getDeclarationModifiers();
                    }
                    scope.problemReporter().javadocInvalidValueReference(alloc.sourceStart, alloc.sourceEnd, scopeModifiers);
                }
                if (!hasProblems && alloc.binding != null && alloc.binding.isValidBinding() && alloc.resolvedType instanceof ReferenceBinding) {
                    ReferenceBinding resolvedType2 = (ReferenceBinding)alloc.resolvedType;
                    this.verifyTypeReference(alloc, alloc.type, scope, source15, resolvedType2, alloc.binding.modifiers);
                }
            }
            else if (reference instanceof JavadocSingleTypeReference && reference.resolvedType != null && reference.resolvedType.isTypeVariable()) {
                scope.problemReporter().javadocInvalidReference(reference.sourceStart, reference.sourceEnd);
            }
            return;
        }
    }
    
    private void resolveParamTags(MethodScope scope, boolean reportMissing, boolean considerParamRefAsUsage) {
        AbstractMethodDeclaration methodDecl = scope.referenceMethod();
        int paramTagsSize = (this.paramReferences == null) ? 0 : this.paramReferences.length;
        if (methodDecl == null) {
            for (int l = 0; l < paramTagsSize; ++l) {
                JavadocSingleNameReference param2 = this.paramReferences[l];
                scope.problemReporter().javadocUnexpectedTag(param2.tagSourceStart, param2.tagSourceEnd);
            }
            return;
        }
        int argumentsSize = (methodDecl.arguments == null) ? 0 : methodDecl.arguments.length;
        if (paramTagsSize == 0) {
            if (reportMissing) {
                for (int k = 0; k < argumentsSize; ++k) {
                    Argument arg2 = methodDecl.arguments[k];
                    scope.problemReporter().javadocMissingParamTag(arg2.name, arg2.sourceStart, arg2.sourceEnd, methodDecl.binding.modifiers);
                }
            }
        }
        else {
            LocalVariableBinding[] bindings = new LocalVariableBinding[paramTagsSize];
            int maxBindings = 0;
            for (int i = 0; i < paramTagsSize; ++i) {
                JavadocSingleNameReference param = this.paramReferences[i];
                param.resolve(scope, true, considerParamRefAsUsage);
                if (param.binding != null && param.binding.isValidBinding()) {
                    boolean found = false;
                    for (int j = 0; j < maxBindings && !found; ++j) {
                        if (bindings[j] == param.binding) {
                            scope.problemReporter().javadocDuplicatedParamTag(param.token, param.sourceStart, param.sourceEnd, methodDecl.binding.modifiers);
                            found = true;
                        }
                    }
                    if (!found) {
                        bindings[maxBindings++] = (LocalVariableBinding)param.binding;
                    }
                }
            }
            if (reportMissing) {
                for (int i = 0; i < argumentsSize; ++i) {
                    Argument arg = methodDecl.arguments[i];
                    boolean found = false;
                    for (int j = 0; j < maxBindings && !found; ++j) {
                        LocalVariableBinding binding = bindings[j];
                        if (arg.binding == binding) {
                            found = true;
                        }
                    }
                    if (!found) {
                        scope.problemReporter().javadocMissingParamTag(arg.name, arg.sourceStart, arg.sourceEnd, methodDecl.binding.modifiers);
                    }
                }
            }
        }
        return;
    }
    
    private void resolveTypeParameterTags(Scope scope, boolean reportMissing) {
        int paramTypeParamLength = (this.paramTypeParameters == null) ? 0 : this.paramTypeParameters.length;
        TypeParameter[] parameters = null;
        TypeVariableBinding[] typeVariables = null;
        int modifiers = -1;
        switch (scope.kind) {
            case 2: {
                AbstractMethodDeclaration methodDeclaration = ((MethodScope)scope).referenceMethod();
                if (methodDeclaration == null) {
                    for (int i = 0; i < paramTypeParamLength; ++i) {
                        JavadocSingleTypeReference param = this.paramTypeParameters[i];
                        scope.problemReporter().javadocUnexpectedTag(param.tagSourceStart, param.tagSourceEnd);
                    }
                    return;
                }
                parameters = methodDeclaration.typeParameters();
                typeVariables = methodDeclaration.binding.typeVariables;
                modifiers = methodDeclaration.binding.modifiers;
                break;
            }
            case 3: {
                TypeDeclaration typeDeclaration = ((ClassScope)scope).referenceContext;
                parameters = typeDeclaration.typeParameters;
                typeVariables = typeDeclaration.binding.typeVariables;
                modifiers = typeDeclaration.binding.modifiers;
                break;
            }
        }
        if (typeVariables == null || typeVariables.length == 0) {
            for (int i2 = 0; i2 < paramTypeParamLength; ++i2) {
                JavadocSingleTypeReference param3 = this.paramTypeParameters[i2];
                scope.problemReporter().javadocUnexpectedTag(param3.tagSourceStart, param3.tagSourceEnd);
            }
            return;
        }
        if (parameters != null) {
            reportMissing = (reportMissing && scope.compilerOptions().sourceLevel >= 3211264L);
            int typeParametersLength = parameters.length;
            if (paramTypeParamLength == 0) {
                if (reportMissing) {
                    for (int i = 0, m = typeParametersLength; i < m; ++i) {
                        scope.problemReporter().javadocMissingParamTag(parameters[i].name, parameters[i].sourceStart, parameters[i].sourceEnd, modifiers);
                    }
                }
            }
            else if (typeVariables.length == typeParametersLength) {
                TypeVariableBinding[] bindings = new TypeVariableBinding[paramTypeParamLength];
                for (int j = 0; j < paramTypeParamLength; ++j) {
                    JavadocSingleTypeReference param2 = this.paramTypeParameters[j];
                    TypeBinding paramBindind = param2.internalResolveType(scope, 0);
                    if (paramBindind != null && paramBindind.isValidBinding()) {
                        if (paramBindind.isTypeVariable()) {
                            if (scope.compilerOptions().reportUnusedParameterIncludeDocCommentReference) {
                                TypeVariableBinding typeVariableBinding = (TypeVariableBinding)paramBindind;
                                typeVariableBinding.modifiers |= 0x8000000;
                            }
                            boolean duplicate = false;
                            for (int k = 0; k < j && !duplicate; ++k) {
                                if (TypeBinding.equalsEquals(bindings[k], param2.resolvedType)) {
                                    scope.problemReporter().javadocDuplicatedParamTag(param2.token, param2.sourceStart, param2.sourceEnd, modifiers);
                                    duplicate = true;
                                }
                            }
                            if (!duplicate) {
                                bindings[j] = (TypeVariableBinding)param2.resolvedType;
                            }
                        }
                        else {
                            scope.problemReporter().javadocUndeclaredParamTagName(param2.token, param2.sourceStart, param2.sourceEnd, modifiers);
                        }
                    }
                }
                for (TypeParameter parameter : parameters) {
                    boolean found = false;
                    for (int l = 0; l < paramTypeParamLength && !found; ++l) {
                        if (TypeBinding.equalsEquals(parameter.binding, bindings[l])) {
                            found = true;
                            bindings[l] = null;
                        }
                    }
                    if (!found && reportMissing) {
                        scope.problemReporter().javadocMissingParamTag(parameter.name, parameter.sourceStart, parameter.sourceEnd, modifiers);
                    }
                }
                for (int j = 0; j < paramTypeParamLength; ++j) {
                    if (bindings[j] != null) {
                        JavadocSingleTypeReference param2 = this.paramTypeParameters[j];
                        scope.problemReporter().javadocUndeclaredParamTagName(param2.token, param2.sourceStart, param2.sourceEnd, modifiers);
                    }
                }
            }
        }
        return;
    }
    
    private void resolveThrowsTags(MethodScope methScope, boolean reportMissing) {
        AbstractMethodDeclaration md = methScope.referenceMethod();
        int throwsTagsLength = (this.exceptionReferences == null) ? 0 : this.exceptionReferences.length;
        if (md == null) {
            for (int i2 = 0; i2 < throwsTagsLength; ++i2) {
                TypeReference typeRef2 = this.exceptionReferences[i2];
                int start = typeRef2.sourceStart;
                int end = typeRef2.sourceEnd;
                if ((typeRef2 instanceof JavadocQualifiedTypeReference)) {
                    start = ((JavadocQualifiedTypeReference)typeRef2).tagSourceStart;
                    end = ((JavadocQualifiedTypeReference)typeRef2).tagSourceEnd;
                }
                else if ((typeRef2 instanceof JavadocSingleTypeReference)) {
                    start = ((JavadocSingleTypeReference)typeRef2).tagSourceStart;
                    end = ((JavadocSingleTypeReference)typeRef2).tagSourceEnd;
                }
                methScope.problemReporter().javadocUnexpectedTag(start, end);
            }
            return;
        }
        int boundExceptionLength = (md.binding == null) ? 0 : md.binding.thrownExceptions.length;
        int thrownExceptionLength = (md.thrownExceptions == null) ? 0 : md.thrownExceptions.length;
        if (throwsTagsLength == 0) {
            if (reportMissing) {
                for (int l = 0; l < boundExceptionLength; ++l) {
                    ReferenceBinding exceptionBinding3 = md.binding.thrownExceptions[l];
                    if (exceptionBinding3 != null && exceptionBinding3.isValidBinding()) {
                        int m;
                        for (m = l; m < thrownExceptionLength && TypeBinding.notEquals(exceptionBinding3, md.thrownExceptions[m].resolvedType); ++m) {}
                        if (m < thrownExceptionLength) {
                            methScope.problemReporter().javadocMissingThrowsTag(md.thrownExceptions[m], md.binding.modifiers);
                        }
                    }
                }
            }
        }
        else {
            int maxRef = 0;
            TypeReference[] typeReferences = new TypeReference[throwsTagsLength];
            for (int i = 0; i < throwsTagsLength; ++i) {
                TypeReference typeRef = this.exceptionReferences[i];
                typeRef.resolve(methScope);
                TypeBinding typeBinding = typeRef.resolvedType;
                if (typeBinding != null && typeBinding.isValidBinding() && typeBinding.isClass()) {
                    typeReferences[maxRef++] = typeRef;
                }
            }
            for (int i = 0; i < boundExceptionLength; ++i) {
                ReferenceBinding exceptionBinding = md.binding.thrownExceptions[i];
                if (exceptionBinding != null) {
                    exceptionBinding = (ReferenceBinding)exceptionBinding.erasure();
                }
                boolean found = false;
                for (int j = 0; j < maxRef && !found; ++j) {
                    if (typeReferences[j] != null) {
                        TypeBinding typeBinding2 = typeReferences[j].resolvedType;
                        if (TypeBinding.equalsEquals(exceptionBinding, typeBinding2)) {
                            found = true;
                            typeReferences[j] = null;
                        }
                    }
                }
                if (!found && reportMissing && exceptionBinding != null && exceptionBinding.isValidBinding()) {
                    int k;
                    for (k = i; k < thrownExceptionLength && TypeBinding.notEquals(exceptionBinding, md.thrownExceptions[k].resolvedType); ++k) {}
                    if (k < thrownExceptionLength) {
                        methScope.problemReporter().javadocMissingThrowsTag(md.thrownExceptions[k], md.binding.modifiers);
                    }
                }
            }
            for (int i = 0; i < maxRef; ++i) {
                TypeReference typeRef = typeReferences[i];
                if (typeRef != null) {
                    boolean compatible = false;
                    for (int j = 0; j < thrownExceptionLength && !compatible; ++j) {
                        TypeBinding exceptionBinding2 = md.thrownExceptions[j].resolvedType;
                        if (exceptionBinding2 != null) {
                            compatible = typeRef.resolvedType.isCompatibleWith(exceptionBinding2);
                        }
                    }
                    if (!compatible && !typeRef.resolvedType.isUncheckedException(false)) {
                        methScope.problemReporter().javadocInvalidThrowsClassName(typeRef, md.binding.modifiers);
                    }
                }
            }
        }
        return;
    }
    
    private void verifyTypeReference(Expression reference, Expression typeReference, Scope scope, boolean source15, ReferenceBinding resolvedType, int modifiers) {
        if (resolvedType.isValidBinding()) {
            int scopeModifiers = -1;
            if (!this.canBeSeen(scope.problemReporter().options.reportInvalidJavadocTagsVisibility, modifiers)) {
                scope.problemReporter().javadocHiddenReference(typeReference.sourceStart, reference.sourceEnd, scope, modifiers);
                return;
            }
            if (reference != typeReference && !this.canBeSeen(scope.problemReporter().options.reportInvalidJavadocTagsVisibility, resolvedType.modifiers)) {
                scope.problemReporter().javadocHiddenReference(typeReference.sourceStart, typeReference.sourceEnd, scope, resolvedType.modifiers);
                return;
            }
            if (resolvedType.isMemberType()) {
                ReferenceBinding topLevelType = resolvedType;
                int packageLength = topLevelType.fPackage.compoundName.length;
                int depth = resolvedType.depth();
                int idx = depth + packageLength;
                char[][] computedCompoundName = new char[idx + 1][];
                computedCompoundName[idx] = topLevelType.sourceName;
                while (topLevelType.enclosingType() != null) {
                    topLevelType = topLevelType.enclosingType();
                    computedCompoundName[--idx] = topLevelType.sourceName;
                }
                int i = packageLength;
                while (--i >= 0) {
                    computedCompoundName[--idx] = topLevelType.fPackage.compoundName[i];
                }
                ClassScope topLevelScope = scope.classScope();
                if (topLevelScope.parent.kind != 4 || !CharOperation.equals(topLevelType.sourceName, topLevelScope.referenceContext.name)) {
                    topLevelScope = topLevelScope.outerMostClassScope();
                    if (typeReference instanceof JavadocSingleTypeReference && ((!source15 && depth == 1) || TypeBinding.notEquals(topLevelType, topLevelScope.referenceContext.binding))) {
                        boolean hasValidImport = false;
                        if (source15) {
                            CompilationUnitScope unitScope = topLevelScope.compilationUnitScope();
                            ImportBinding[] imports = unitScope.imports;
                        Label_0467:
                            for (int length = (imports == null) ? 0 : imports.length, j = 0; j < length; ++j) {
                                char[][] compoundName = imports[j].compoundName;
                                int compoundNameLength = compoundName.length;
                                if ((imports[j].onDemand && compoundNameLength == computedCompoundName.length - 1) || compoundNameLength == computedCompoundName.length) {
                                    int k = compoundNameLength;
                                    while (--k >= 0 && CharOperation.equals(imports[j].compoundName[k], computedCompoundName[k])) {
                                        if (k == 0) {
                                            hasValidImport = true;
                                            ImportReference importReference = imports[j].reference;
                                            if (importReference != null) {
                                                importReference.bits |= 0x2;
                                                break Label_0467;
                                            }
                                            break Label_0467;
                                        }
                                        else {
                                            continue;
                                        }
                                    }
                                }
                            }
                            if (!hasValidImport) {
                                if (scopeModifiers == -1) {
                                    scopeModifiers = scope.getDeclarationModifiers();
                                }
                                scope.problemReporter().javadocInvalidMemberTypeQualification(typeReference.sourceStart, typeReference.sourceEnd, scopeModifiers);
                            }
                        }
                        else {
                            if (scopeModifiers == -1) {
                                scopeModifiers = scope.getDeclarationModifiers();
                            }
                            scope.problemReporter().javadocInvalidMemberTypeQualification(typeReference.sourceStart, typeReference.sourceEnd, scopeModifiers);
                            return;
                        }
                    }
                }
                if (typeReference instanceof JavadocQualifiedTypeReference && !scope.isDefinedInSameUnit(resolvedType)) {
                    char[][] typeRefName = ((JavadocQualifiedTypeReference)typeReference).getTypeName();
                    int skipLength = 0;
                    if (topLevelScope.getCurrentPackage() == resolvedType.getPackage() && typeRefName.length < computedCompoundName.length) {
                        skipLength = resolvedType.fPackage.compoundName.length;
                    }
                    boolean valid = true;
                    if (typeRefName.length == computedCompoundName.length - skipLength) {
                        int l = 0;
                        while (l < typeRefName.length) {
                            if (!CharOperation.equals(typeRefName[l], computedCompoundName[l + skipLength])) {
                                valid = false;
                                break;
                            }
                            ++l;
                        }
                    }
                    else {
                        valid = false;
                    }
                    if (!valid) {
                        if (scopeModifiers == -1) {
                            scopeModifiers = scope.getDeclarationModifiers();
                        }
                        scope.problemReporter().javadocInvalidMemberTypeQualification(typeReference.sourceStart, typeReference.sourceEnd, scopeModifiers);
                        return;
                    }
                }
            }
            if (scope.referenceCompilationUnit().isPackageInfo() && typeReference instanceof JavadocSingleTypeReference && resolvedType.fPackage.compoundName.length > 0) {
                scope.problemReporter().javadocInvalidReference(typeReference.sourceStart, typeReference.sourceEnd);
                return;
            }
        }
    }
    
    @Override
    public void traverse(ASTVisitor visitor, BlockScope scope) {
        if (visitor.visit(this, scope)) {
            if (this.paramReferences != null) {
                for (int i = 0, length = this.paramReferences.length; i < length; ++i) {
                    this.paramReferences[i].traverse(visitor, scope);
                }
            }
            if (this.paramTypeParameters != null) {
                for (int i = 0, length = this.paramTypeParameters.length; i < length; ++i) {
                    this.paramTypeParameters[i].traverse(visitor, scope);
                }
            }
            if (this.returnStatement != null) {
                this.returnStatement.traverse(visitor, scope);
            }
            if (this.exceptionReferences != null) {
                for (int i = 0, length = this.exceptionReferences.length; i < length; ++i) {
                    this.exceptionReferences[i].traverse(visitor, scope);
                }
            }
            if (this.seeReferences != null) {
                for (int i = 0, length = this.seeReferences.length; i < length; ++i) {
                    this.seeReferences[i].traverse(visitor, scope);
                }
            }
        }
        visitor.endVisit(this, scope);
    }
    
    public void traverse(ASTVisitor visitor, ClassScope scope) {
        if (visitor.visit(this, scope)) {
            if (this.paramReferences != null) {
                for (int i = 0, length = this.paramReferences.length; i < length; ++i) {
                    this.paramReferences[i].traverse(visitor, scope);
                }
            }
            if (this.paramTypeParameters != null) {
                for (int i = 0, length = this.paramTypeParameters.length; i < length; ++i) {
                    this.paramTypeParameters[i].traverse(visitor, scope);
                }
            }
            if (this.returnStatement != null) {
                this.returnStatement.traverse(visitor, scope);
            }
            if (this.exceptionReferences != null) {
                for (int i = 0, length = this.exceptionReferences.length; i < length; ++i) {
                    this.exceptionReferences[i].traverse(visitor, scope);
                }
            }
            if (this.seeReferences != null) {
                for (int i = 0, length = this.seeReferences.length; i < length; ++i) {
                    this.seeReferences[i].traverse(visitor, scope);
                }
            }
        }
        visitor.endVisit(this, scope);
    }
}
